@@ -10,8 +10,6 @@ from i2.deco import (
 
 from i2.signatures import (
     Sig,
-    Command,
-    extract_commands,
     call_forgivingly,
     call_somewhat_forgivingly,
     name_of_obj,
@@ -32,3 +30,16 @@ from i2.errors import (
     InterruptWithBlock,
     HandleExceptions,
 )
+
+from i2.wrapper import (
+    wrap,
+    Wrap,
+    Wrapx,
+    Ingress,
+    include_exclude,
+    rm_params,
+    partialx,
+    map_names,
+    func_to_method_func,
+    bind_funcs_object_attrs,
+)
