@@ -13,318 +13,71 @@ Both in the code and in the docs, we'll use short hands for parameter (argument)
 """
 import pytest
 from functools import reduce
-from i2.signatures import _empty
+from typing import Any
+
 from i2.signatures import *
 from i2.signatures import normalized_func
 
-from typing import Iterable, Union, Callable, List, Any
-from inspect import Parameter, Signature
-from functools import singledispatch
+from i2.tests.util import sig_to_inputs, trace_call
 
 
-# ------------------------------------------------------------------------------------
-# Test utils
+def test_sig_wrap_edge_cases():
+    """Tests some edge cases involving simultaneous changes of defaults and kinds.
 
-ParameterAble = Union[int, Parameter, str]
-ParamsAble_ = Union[ParamsAble, str, List[int]]
+    Current Sig.wrap design allows you to change defaults, and this will have the
+    effect of changing the ``__defaults__`` and ``__kwdefaults__`` of the wrapped
+    function.
 
+    But, when we change the default of a parameter and change it's kind at the same
+    time, bad things happen (see https://github.com/i2mint/i2/issues/16).
 
-def _to_params(params: ParamsAble_):
-    """
+    So `Sig.wrap` makes a few checks and raises an error if it's thinks it's not
+    safe to do the wrapping.
 
-    >>> str(Sig(_to_params([0, 0, 1, 1, 1, 2, 3, 4])))
-    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
+    Some of these facts may change with new designs. When this is the case,
+    this test should be invalidated."""
 
-    >>> str(Sig(_to_params("00111234")))
-    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
-    """
-    if isinstance(params, (Callable, Signature)):
-        yield from Sig(params).params
-    else:
-        for i, spec in enumerate(params):
-            if isinstance(spec, int):
-                kind = spec
-                yield Parameter(f'a{kind}{i}', kind=kind)
-            elif isinstance(spec, Parameter):
-                param = spec
-                yield param
-            elif isinstance(spec, str) and spec.isnumeric():
-                kind = int(spec)
-                yield Parameter(f'a{kind}{i}', kind=kind)
-            else:
-                try:
-                    yield ensure_param(spec)
-                except Exception:
-                    raise TypeError(
-                        f"Don't know how to handle this type of obj: {spec}"
-                    )
+    # Non edge-case tests:
 
+    def foo(x, y, z=0):
+        return x + y * z
 
-def _params_to_arg_name_and_val(params: ParamsAble_):
-    """
+    assert foo.__defaults__ == (0,)
+    assert foo(1, 2) == 1
 
-    >>> assert dict(_params_to_arg_name_and_val(_to_params("00111234"))) == {
-    ...     "a00": 0,
-    ...     "a01": 1,
-    ...     "a12": 2,
-    ...     "a13": 3,
-    ...     "a14": 4,
-    ...     "a25": (5, -5),
-    ...     "a36": 6,
-    ...     "a47": {"a47": 7, "a47_": -7},
-    ... }
-    """
-    params = _to_params(params)
-    for i, param in enumerate(params):
-        if param.kind == Parameter.VAR_POSITIONAL:
-            val = (i, -i)
-        elif param.kind == Parameter.VAR_KEYWORD:
-            val = {param.name: i, param.name + '_': -i}
-        else:
-            val = i
-        yield (param.name, val)
+    @Sig(lambda x, y, z=3: None)
+    def foo(x, y, z=0):
+        return x + y * z
 
+    assert foo(1, 2) == 7
+    #    works because Sig also changed __defaults__:
+    assert foo.__defaults__ == (3,)
 
-assert dict(_params_to_arg_name_and_val(_to_params('00111234'))) == {
-    'a00': 0,
-    'a01': 1,
-    'a12': 2,
-    'a13': 3,
-    'a14': 4,
-    'a25': (5, -5),
-    'a36': 6,
-    'a47': {'a47': 7, 'a47_': -7},
-}
+    @Sig(lambda x, y, *, z=3: None)
+    def foo(x, y, *, z=0):
+        return x + y * z
 
+    assert foo(1, 2) == 7
+    #    works because Sig also changed __defaults__ and __kwdefaults__:
+    assert foo.__defaults__ == ()
+    assert foo.__kwdefaults__ == {'z': 3}
 
-def inject_defaults(params: ParamsAble_, defaults: dict):
-    """Yields params with defaults ({argname: default_val,...}) edited.
+    # The following (where we go from a (same kind) param not having a default,
+    # to having one, also work
 
-    >>> assert (
-    ...     str(
-    ...         Sig(
-    ...             inject_defaults(
-    ...                 _to_params("00111234"), defaults={"a14": 40, "a36": 60}
-    ...             )
-    ...         )
-    ...     )
-    ...     == "(a00, a01, /, a12, a13, a14=40, *a25, a36=60, **a47)"
-    ... )
-    """
-    for param in _to_params(params):
-        if param.name in defaults:
-            yield param.replace(default=defaults[param.name])
-        else:
-            yield param
+    @Sig(lambda x, y, z=3: None)
+    def foo(x, y, z):
+        return x + y * z
 
+    assert foo(1, 2) == 7
+    assert foo(1, 2, z=10) == 21
 
-assert (
-    str(Sig(inject_defaults(_to_params('00111234'), defaults={'a14': 40, 'a36': 60})))
-    == '(a00, a01, /, a12, a13, a14=40, *a25, a36=60, **a47)'
-)
+    @Sig(lambda x, y, *, z=3: None)
+    def foo(x, y, *, z):
+        return x + y * z
 
-
-def _str_of_call_args(_call_kwargs: dict):
-    return ', '.join(f'{k}={v}' for k, v in _call_kwargs.items())
-
-
-def mk_func_from_params(
-    params: ParamsAble = '00111234',
-    *,
-    defaults=None,
-    name=None,
-    callback: Callable[[dict], Any] = _str_of_call_args,
-):
-    """Make a function (that actually returns something based on args) from params.
-
-    :param params: params (arguments) of the function (can be expressed in many ways!)
-    :param defaults: Optional {argname: default,...} dict to inject defaults
-    :param name: Optional name to give the function
-    :param callback: The function defining what the function actually does.
-        Must be a function taking a single dict input encapsulating the all arguments.
-        The default will return a string representation of this dict.
-    :return: A function with the specified params, returning a string of it's (call) args
-
-    There's many ways you can express the `params` input.
-    Any of the ways understood by the `signatures.ensure_params` function, for one;
-    plus a few more.
-
-    One nice way to express the params is through an actual function.
-    Note that the code of the function isn't even looked out.
-    Only it's signature is taken into consideration.
-    The returned function will have the same signature.
-    Instead, the callback function will be acalled on the infered _call_kwargs
-    dict of {argname: argval} pairs.
-    The default callaback is a string exhibiting these argname/argval pairs.
-
-    >>> f = mk_func_from_params(lambda x, /, y, *, z: None)
-    >>> print(f"{f.__name__}{Sig(f)}")
-    f(x, /, y, *, z)
-    >>> f(1, 2, z=3)
-    'x=1, y=2, z=3'
-    >>> f(1, y=2, z=3)
-    'x=1, y=2, z=3'
-    >>> f = mk_func_from_params(lambda x, /, y=42, *, z='ZZZ': None)
-    >>> print(f"{f.__name__}{Sig(f)}")
-    f(x, /, y=42, *, z='ZZZ')
-    >>> f(3.14)
-    'x=3.14, y=42, z=ZZZ'
-
-    If you're not interested in having that level of control, but are just
-    interested in the number and kinds of the arguments, you can specify only that;
-    a sequence of kinds.
-    These must be a non-decreasing sequence of integers between
-    0 and 4 inclusive. These integers represent kinds of parameters.
-    See https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
-    to see what each integer value means.
-    You can also specify this integer sequence as a single string, as shown below.
-
-    >>> f = mk_func_from_params(params="00111234")
-    >>> str(Sig(f))
-    '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
-    >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
-    "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
-    >>> f(0, 1, 2, a13=3, a14=4, a36=6)
-    'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'
-
-    What just happened?
-    Well, `params="00111234"` was transformed to `params=[0, 0, 1, 1, 1, 2, 3, 4]`,
-    which was transformed to a list of the same size, using
-
-    Now, if you really want full control over those params, you can specify them
-    completely using the `inspect.Parameter` class.
-    You can also decide what level of control you want, and mix and match all kinds of
-    specifications, as below.
-
-    >>> from inspect import Parameter
-    >>> f = mk_func_from_params([
-    ...     0,
-    ...     'blah',
-    ...     Parameter(name='hello',
-    ...               kind=Parameter.POSITIONAL_OR_KEYWORD,
-    ...               default='world')
-    ... ])
-    >>> print(f"{f.__name__}{Sig(f)}")
-    f(a00, /, blah, hello='world')
-    >>> assert f(11, 22) == 'a00=11, blah=22, hello=world'
-
-    """
-    params = _to_params(params)
-    params = inject_defaults(params, defaults=defaults or {})
-    sig = Sig(params)
-
-    @sig
-    def arg_str_func(*args, **kwargs):
-        _call_kwargs = sig.kwargs_from_args_and_kwargs(
-            args, kwargs, apply_defaults=True
-        )
-        return callback(_call_kwargs)
-
-    arg_str_func.__name__ = name or 'f' + ''.join(str(p.kind) for p in params)
-
-    return arg_str_func
-
-
-def mk_func_inputs_for_params(params: ParamsAble_, param_to_input):
-    pass
-
-
-#
-# @mk_func_from_params.register
-# def mk_func_from_params(params: Iterable[int], defaults=None, name=None):
-#     """
-#
-#     :param kinds:
-#     :param defaults:
-#     :return:
-#
-#     Make a sequence of kinds (must be a non-decreasing sequence of integers between
-#     0 and 4 inclusive. These integers represent kinds of parameters.
-#     See https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
-#     to see what each integer value means.
-#
-#     >>> kinds = list(map(int, "00111234"))
-#     >>> kinds
-#     [0, 0, 1, 1, 1, 2, 3, 4]
-#
-#     Note: `kinds_to_arg_str_func` also works directly with strings such as "00111234".
-#
-#     Now tell `kinds_to_arg_str_func` to make a function with those kinds.
-#
-#     >>> f = kinds_to_arg_str_func(kinds)
-#     >>> str(Sig(f))
-#     '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
-#     >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
-#     "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
-#     >>> f(0, 1, 2, a13=3, a14=4, a36=6)
-#     'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'
-#     """
-#     kinds = params
-#     params = inject_defaults(_to_params(kinds), defaults=defaults or {})
-#     name = name or "f" + "".join(map(str, kinds))
-#
-#     return mk_func_from_params(params, defaults, name)
-
-#
-#
-# @mk_func_from_params.register
-# def _(kinds: str):
-#     """
-#     >>> f = kinds_to_arg_str_func("00111234")
-#     >>> str(Sig(f))
-#     '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
-#     >>> f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
-#     "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
-#     >>> f(0, 1, 2, a13=3, a14=4, a36=6)
-#     'a00=0, a01=1, a12=2, a13=3, a14=4, a25=(), a36=6, a47={}'
-#
-#     """
-#     return mk_func_from_params(_kinds_str_to_int_list(kinds))
-#
-#
-# f = mk_func_from_params("00111234")
-# assert str(Sig(f)) == "(a00, a01, /, a12, a13, a14, *a25, a36, **a47)"
-# assert (
-#     f(0, 1, 2, 3, 4, 5, -5, a36=6, a47={"a47": 7, "a47_": -7})
-#     == "a00=0, a01=1, a12=2, a13=3, a14=4, a25=(5, -5), a36=6, a47={'a47': {'a47': 7, 'a47_': -7}}"
-# )
-
-
-empty = _empty
-
-mappingproxy = type(Signature().parameters)
-
-
-def trace_call(func, local_vars, name=None):
-    if name is None:
-        name = func.__name__
-    return (
-        f'{name}('
-        + ', '.join(f'{argname}={local_vars[argname]}' for argname in Sig(func).names)
-        + ')'
-    )
-
-
-# class KeywordArg(dict):
-#     """Just to mark a dict as a keyword argument"""
-#
-#
-# def _separate_pk_arguments_into_positional_and_keyword(pka):
-#     args = []
-#     kwargs = {}
-#     pka_iter = iter(pka)
-#     for a in pka_iter:
-#         if not isinstance(a, KeywordArg):
-#             args.append(a)
-#         else:
-#             kwargs.update(dict(a))
-#     for a in pka_iter:
-#         kwargs.update(dict(a))
-#
-#     return args, kwargs
-
-# ------------------------------------------------------------------------------------
-# And finally... the tests
+    assert foo(1, 2) == 7
+    assert foo(1, 2, z=10) == 21
 
 
 def test_tuple_the_args():
@@ -699,3 +452,165 @@ def update_signature_with_signatures_from_funcs(*funcs, priority: str = 'last'):
         raise ValueError("priority should be 'last' or 'first'")
 
     return transform_signature
+
+
+@pytest.mark.parametrize(
+    'sig_spec',
+    [
+        ('(po, /)'),
+        ('(po=0, /)'),
+        ('(pk)'),
+        ('(pk=0)'),
+        ('(*, ko)'),
+        ('(*, ko=0)'),
+        ('(po, /, pk, *, ko)'),
+        ('(po=0, /, pk=0, *, ko=0)'),
+        ('(*args)'),
+        ('(**kwargs)'),
+        ('(*args, **kwargs)'),
+        ('(po, /, pk, *args, ko)'),
+        ('(po=0, /, pk=0, *args, ko=0)'),
+        ('(po, /, pk, *, ko, **kwargs)'),
+        ('(po=0, /, pk=0, *, ko=0, **kwargs)'),
+        ('(po, /, pk, *args, ko, **kwargs)'),
+        ('(po=0, /, pk=0, *args, ko=0, **kwargs)'),
+        ('(po1, po2, /)'),
+        ('(pk1, pk2)'),
+        ('(*, ko1, ko2)'),
+        ('(po1, po2, /, pk1, pk2, *, ko1, ko2)'),
+        ('(po1, po2, /, pk1, pk2, *args, ko1, ko2, **kwargs)'),
+        ('(po1=0, po2=0, /, pk1=0, pk2=0, *args, ko1=0, ko2=0, **kwargs)'),
+    ],
+)
+def test_call_forgivingly(sig_spec):
+    sig = Sig(sig_spec)
+
+    @sig
+    def foo(*args, **kwargs):
+        return args, kwargs
+
+    def validate_call_forgivingly(*args, **kwargs):
+        expected_output_kwargs = (
+            kwargs
+            if VK in sig.kinds.values()
+            else {k: v for k, v in kwargs.items() if k in sig}
+        )
+        pk_in_kwargs_count = sum(
+            kind == PK
+            for param, kind in sig.kinds.items()
+            if param in expected_output_kwargs
+        )
+        expected_output_args_count = (
+            len(args)
+            if VP in sig.kinds.values()
+            else sum(kind <= PK for kind in sig.kinds.values()) - pk_in_kwargs_count
+        )
+        expected_output_args = args[:expected_output_args_count]
+        expected_output = (expected_output_args, expected_output_kwargs)
+        output = call_forgivingly(foo, *args, **kwargs)
+        # print()
+        # print(args, kwargs)
+        # print(expected_output)
+        # print(output)
+        assert output == expected_output
+
+    for args, kwargs in sig_to_inputs(sig, ignore_variadics=True):
+        kwargs = dict(kwargs, **{'some': 'extra', 'added': 'kwargs'})
+        po_pk_count = sum(kind <= PK for kind in sig.kinds.values())
+        if len(args) == po_pk_count:
+            args = args + ('some', 'extra', 'args')
+
+        validate_call_forgivingly(*args, **kwargs)
+
+
+@pytest.mark.parametrize(
+    'sig_spec1, sig_spec2',
+    [
+        ('()', '(a)'),
+        ('()', '(a=0)'),
+        ('(a, /, *, c)', '(a, /, b, *, c)'),
+        ('(a, /, *, c)', '(a, /, b=0, *, c)'),
+        ('(a, /, b)', '(a, /, b, *, c)'),
+        ('(a, /, b)', '(a, /, b, *, c=0)'),
+        ('(a, /, b, *, c)', '(*args)'),
+        ('(a, /, b, *, c)', '(**kwargs)'),
+        ('(a, /, b, *, c)', '(*args, **kwargs)'),
+        ('(a, /, b, *, c)', '(a, /, b, *args, c, **kwargs)'),
+        ('(a, /, b, *, c)', '(a, b, c)'),
+        ('(a, /, b, *, c)', '(a, b, /, *, c)'),
+        ('(a, /, b, *, c)', '(a, /, *, b, c)'),
+        ('(a, /, b, *, c)', '(a, b, /, c)'),
+        ('(a, /, b, *, c)', '(a, *, b, c)'),
+        ('(a, /, b, *, c)', '(x, /, b, *, c)',),
+        ('(a, /, b, *, c)', '(a, /, x, *, c)',),
+        ('(a, /, b, *, c)', '(a, /, b, *, x)',),
+        ('(a, /, b, *, c)', '(a=0, b=0, c=0)'),
+        ('(a=0, /, b=0, *, c=0)', '(a, b, c)'),
+        ('(a, b, /, c, d, *, e, f)', '(b, a, /, d, c, *, f, e)'),
+        ('(a, b, /, c, d, *, e, f)', '(a, c, /, b, e, *, d, f)',),
+        # (
+        #     '()',
+        #     '(*args)'
+        # ),
+        # (
+        #     '()',
+        #     '(**kwargs)'
+        # ),
+        # (
+        #     '()',
+        #     '(*args, **kwargs)'
+        # ),
+        # (
+        #     '(*args)',
+        #     '()'
+        # ),
+        # (
+        #     '(*args)',
+        #     '(**kwargs)'
+        # ),
+        # (
+        #     '(*args)',
+        #     '(*args, **kwargs)'
+        # ),
+        # (
+        #     '(**kwargs)',
+        #     '()'
+        # ),
+        # (
+        #     '(**kwargs)',
+        #     '(*args)'
+        # ),
+        # (
+        #     '(**kwargs)',
+        #     '(*args, **kwargs)'
+        # ),
+        # (
+        #     '(*args, **kwargs)',
+        #     '(*args, **kwargs)'
+        # ),
+    ],
+)
+def test_signature_comparison(sig_spec1, sig_spec2):
+    sig1 = Sig(sig_spec1)
+    sig2 = Sig(sig_spec2)
+    is_compatible = sig1 <= sig2
+
+    exec_env = dict()
+    f_def = f'def f{sig_spec2}: pass'
+    exec(f_def, exec_env)
+    foo = exec_env['f']
+
+    # @sig2
+    # def foo(*args, **kwargs):
+    #     pass
+
+    for args, kwargs in sig_to_inputs(sig1, ignore_variadics=True):
+        try:
+            foo(*args, **kwargs)
+        except TypeError:
+            if is_compatible:
+                raise
+            else:
+                return
+    if not is_compatible:
+        raise AssertionError('sig1 is compatible with sig2, when it should not.')
