@@ -65,11 +65,26 @@ How the ``Ingress`` class (ingress templated function maker) works:
 
 from functools import wraps, partial
 from inspect import Parameter, signature
-from typing import Mapping, Callable, Optional
+from typing import (
+    Mapping,
+    Callable,
+    Optional,
+    Union,
+    Iterable,
+    Sequence,
+    NewType,
+    Dict,
+    Tuple,
+)
+
 from types import MethodType
 
-from i2.signatures import Sig
+from i2.signatures import Sig, name_of_obj
 from i2.multi_object import Pipe
+from i2.deco import double_up_as_factory
+
+# ---------------------------------------------------------------------------------------
+# Wrap
 
 empty = Parameter.empty
 OuterKwargs = dict
@@ -110,7 +125,57 @@ class MakeFromFunc:
         return self.func_to_obj(func)
 
 
-class Wrap:
+# from wrapt import decorator
+# TODO: Continue factoring out Wrap and Wrapx code
+class _Wrap:
+    _init_args = ()
+    _init_kwargs = ()
+
+    def __init__(self, func, *args, **kwargs):
+        self._init_args = (func, *args)
+        self._init_kwargs = kwargs
+        # wraps(func)(self) is there to copy over to self anything that func may
+        # have had. It should be before anything else so it doesn't overwrite stuff
+        # that we may add to self in init (like .func for example!)
+        wraps(func)(self)  # TODO: should we really copy everything by default?
+        if name := kwargs.get('name', None) is not None:
+            self.__name__ = name
+        self.func = func  # Note: overwrites self.func that wraps MAY have inserted
+        self.__wrapped__ = func
+        # TODO: Pros and cons analysis of pointing __wrapped__ to func. partial uses
+        #  .func, but wraps looks for __wrapped__
+
+    def __reduce__(self):
+        """reduce is meant to control how things are pickled/unpickled"""
+        return type(self), self._init_args, dict(self._init_kwargs)
+
+    def __get__(self, instance, owner):
+        """This method allows things to work well when we use Wrap object as method"""
+        return MethodType(self, instance)
+
+    def __repr__(self):
+        # TODO: Replace i2.Wrap with dynamic (Wrap or Wrapx)
+        name = getattr(self, '__name__', None) or 'Wrap'
+        return f'<i2.Wrap {name}{signature(self)}>'
+
+    # TODO: Don't know exactly what I'm doing below. Review with someone!
+    def __set_name__(self, owner, name):
+        """So that name of function is passed on to method when assigning to attribute
+        That is, doing ``method = Wrap(func)`` in a class definition"""
+        # TODO: Look into sanity of mutating the name and other ways to achieve same
+        self.__name__ = name
+
+    def __set__(self, instance, value):
+        instance.__dict__[self.__name__] = value
+
+    # To get help(instance.method) to work!
+    # TODO: Does this have undesirable side effects?
+    @property
+    def __code__(self):
+        return self.func.__code__
+
+
+class Wrap(_Wrap):
     """A function wrapper with interface modifiers.
 
     :param func: The wrapped function
@@ -119,6 +184,7 @@ class Wrap:
         wrapped function.
     :param egress: The outgoing data transformer. It also takes precedence over the
         wrapped function to determine the return annotation of the ``Wrap`` instance
+    :param name: Name to give the wrapper (will use wrapped func name by default)
     :return: A callable instance wrapping ``func``
 
     Some examples:
@@ -184,13 +250,13 @@ class Wrap:
     >>> def func(times, string):
     ...     return times * string
     ...
-    >>> wrapped_func = wrap(func, ingress)
+    >>> wrapped_func = wrap(func, ingress=ingress)
     >>> assert wrapped_func(2, "world! ", "Hi") == "Hi world! Hi world! Hi world! "
     >>>
     >>> wrapped_func = wrap(func, egress=len)
     >>> assert wrapped_func(2, "co") == 4 == len("coco") == len(func(2, "co"))
     >>>
-    >>> wrapped_func = wrap(func, ingress, egress=len)
+    >>> wrapped_func = wrap(func, ingress=ingress, egress=len)
     >>> assert (
     ...     wrapped_func(2, "world! ", "Hi")
     ...     == 30
@@ -211,18 +277,8 @@ class Wrap:
 
     """
 
-    def __init__(self, func, ingress=None, egress=None):
-        # wraps(func)(self) is there to copy over to self anything that func may
-        # have had. It should be before anything else so it doesn't overwrite stuff
-        # that we may add to self in init (like .func for example!)
-        wraps(func)(self)  # TODO: should we really copy everything by default?
-
-        self.func = func  # Note: overwrites self.func that wraps MAY have inserted
-
-        # remember the actual value of ingress and egress (for reduce to reproduce)
-        self._ingress = ingress
-        self._egress = egress
-
+    def __init__(self, func, ingress=None, egress=None, *, name=None):
+        super().__init__(func, ingress, egress, name=name)
         ingress_sig = Sig(func)
 
         if ingress is None:
@@ -251,22 +307,16 @@ class Wrap:
                 return_annotation = egress_return_annotation
 
         self.__signature__ = Sig(ingress_sig, return_annotation=return_annotation)
-        self.__wrapped__ = func
-        # TODO: Pros and cons analysis of pointing __wrapped__ to func. partial uses
-        #  .func, but wraps looks for __wrapped__
 
     def __call__(self, *ingress_args, **ingress_kwargs):
         func_args, func_kwargs = self.ingress(*ingress_args, **ingress_kwargs)
         return self.egress(self.func(*func_args, **func_kwargs))
 
-    def __reduce__(self):
-        return type(self), (self.func, self._ingress, self._egress)
 
-    def __get__(self, instance, owner):
-        return MethodType(self, instance)
-
-
-def wrap(func, ingress=None, egress=None):
+@double_up_as_factory
+def wrap(
+    func=None, *, ingress=None, egress=None, caller=None, name=None, dflt_wrap=Wrap
+):
     """Wrap a function, optionally transforming interface, input and output.
 
     :param func: The wrapped function
@@ -325,7 +375,7 @@ def wrap(func, ingress=None, egress=None):
     Now we can use these ingress and egress functions to get the version of ``f`` of
     our dreams:
 
-    >>> h = wrap(f, ingress, egress)
+    >>> h = wrap(f, ingress=ingress, egress=egress)
 
     Let's see what the signature of our new function looks like:
 
@@ -346,14 +396,43 @@ def wrap(func, ingress=None, egress=None):
 
     >>> assert h(0) == h(0, 1) == h(0, 1, 2) == 0 + 1 * 2 ** 10 == 2 ** 10 == 1024
 
+    Note that ``wrap`` can also be used as a decorator "factory", for instance to
+    wrap functions at definition time, and if we change ``caller`` it will automatically
+    use ``Wrapx`` instead of ``Wrap`` to wrap the function.
+
+    >>> def iterize(func, args, kwargs):
+    ...     first_arg_val = next(iter(kwargs.values()))
+    ...     return list(map(func, first_arg_val))
+    >>>
+    >>> @wrap(caller=iterize)
+    ... def func(x, y=2):
+    ...     return x + y
+    ...
+    >>> isinstance(func, Wrapx)
+    True
+    >>> func([1, 2, 3, 4])
+    [3, 4, 5, 6]
+
     For more examples, see also the
 
     .. seealso::
 
         ``Wrap`` class.
+        ``Wrapx`` class.
 
     """
-    return Wrap(func, ingress, egress)
+    if _should_use_wrapx(func, ingress, egress, caller):
+        return Wrapx(func, ingress, egress, caller=caller, name=name)
+    else:
+        return dflt_wrap(func, ingress, egress, name=name)
+
+
+# TODO: Add conditions on egress to route to Wrapx when complex egress
+def _should_use_wrapx(func, ingress, egress, caller):
+    if caller is not None:
+        return True
+    else:
+        return False
 
 
 def append_empty_args(func):
@@ -461,7 +540,7 @@ class Ingress:
     >>> ingress = Ingress(signature(f), kwargs_trans=kwargs_trans)
     >>> assert ingress(2, x=3, y=4) == ((4,), {'x': 6, 'y': 7, 'z': 3})
     >>>
-    >>> wrapped_f = wrap(f, ingress)
+    >>> wrapped_f = wrap(f, ingress=ingress)
     >>> assert wrapped_f(2, x=3, y=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'
 
 
@@ -486,9 +565,13 @@ class Ingress:
     ... )
     >>> assert ingress(2, x=3, you=4) == ((4,), {'x': 6, 'y': 7, 'z': 3})
     >>>
-    >>> wrapped_f = wrap(f, ingress)
+    >>> wrapped_f = wrap(f, ingress=ingress)
     >>> assert wrapped_f(2, x=3, you=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'
 
+    A convenience method allows to do the same with the ingress instance itself:
+
+    >>> wrapped_f = ingress.wrap(f)
+    >>> assert wrapped_f(2, x=3, you=4) == '(w:=4) + (x:=6) * (y:=7) ** (z:=3) == 2062'
     """
 
     def __init__(
@@ -542,7 +625,7 @@ class Ingress:
             **self.kwargs_trans(func_kwargs),  # change those that kwargs_trans desires
         )
 
-        # Return an (args,kwargs) pair that respects the inner function's
+        # Return an (args, kwargs) pair that respects the inner function's
         # argument kind restrictions.
         # Note: Originally was with (default) allow_excess=False. Changed to True to
         #       allow more flexibility in outer sig. But is this sane? Worth it?
@@ -551,13 +634,37 @@ class Ingress:
             func_kwargs, apply_defaults=True, allow_excess=True
         )
 
+    def __repr__(self):
+        return f'Ingress signature: {signature(self)}'
+
+    def wrap(self, func: Callable, egress=None, *, name=None) -> Wrap:
+        """Convenience method to wrap a function with the instance ingress.
+        ``ingress.wrap(func,...)`` equivalent to ``Wrap(func, ingress, ...)``
+        """
+        return Wrap(func, ingress=self, egress=egress, name=name)
+
     @classmethod
-    def name_map(cls, wrapped, **new_names):
-        """Change argument names"""
+    def name_map(cls, wrapped, **old_to_new_name):
+        """Change argument names.
+
+        >>> def f(w, /, x: float, y=2, *, z: int = 3):
+        ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
+        >>> ingress = Ingress.name_map(f, w='DoubleYou', z='Zee')
+        >>> ingress
+        Ingress signature: (DoubleYou, /, x: float, y=2, *, Zee: int = 3)
+        >>> wrapped_f = ingress.wrap(f)
+        >>> wrapped_f(1, 2, y=3, Zee=4)
+        '(w:=1) + (x:=2) * (y:=3) ** (z:=4) == 163'
+
+        """
+        new_to_old_name = {v: k for k, v in old_to_new_name.items()}
+        assert len(new_to_old_name) == len(
+            old_to_new_name
+        ), f'Inversion is not possible since {old_to_new_name=} has duplicate values.'
         return cls(
             wrapped,
-            partial(Pipe(items_with_mapped_keys, dict), key_mapper=new_names),
-            Sig(wrapped).ch_names(**new_names),
+            partial(Pipe(items_with_mapped_keys, dict), key_mapper=new_to_old_name),
+            Sig(wrapped).ch_names(**old_to_new_name),
         )
 
     #     @classmethod
@@ -574,9 +681,24 @@ class Ingress:
 
 
 def items_with_mapped_keys(d: dict, key_mapper):
+    """Transform dict keys. More precisely yield (new_k,v) pairs from a key mapper dict.
+
+    :param d: src dict
+    :param key_mapper: {old_name: new_name, ...} mapping
+    :return: generator of (new_name, value) pairs
+
+    Often used in conjunction with dict:
+
+    >>> dict(items_with_mapped_keys(
+    ...     {'a': 1, 'b': 2, 'c': 3, 'd': 4},
+    ...     {'a': 'Ay', 'd': 'Dee'})
+    ... )
+    {'Ay': 1, 'b': 2, 'c': 3, 'Dee': 4}
+
+    """
     for k, v in d.items():
-        # key_mapper.get(k, k) will give the new key name if present, else will use
-        # the old
+        # key_mapper.get(k, k) will give the new key name if present,
+        # else will use the old
         yield key_mapper.get(k, k), v
 
 
@@ -623,6 +745,7 @@ def _handle_ingress_class_inputs(
     return inner_sig, kwargs_trans, outer_sig
 
 
+# TODO: See what this adds over ``Ingress`` class. Consider merging or reusing.
 class InnerMapIngress:
     """A class to help build ingresses systematically by mapping the inner signature.
 
@@ -849,6 +972,8 @@ def mk_ingress_from_name_mapper(func, name_mapper: Mapping, *, conserve_kind=Fal
     return ArgNameMappingIngress(func, conserve_kind=conserve_kind, **name_mapper)
 
 
+# TODO: Replace with i2.wrapper instead
+# TODO: Make sure VARIADICs are handled properly, or error raised if present
 def nice_kinds(func):
     """Wraps the func so it will only have POSITIONAL_OR_KEYWORD argument kinds.
 
@@ -872,6 +997,8 @@ def nice_kinds(func):
     >>> isinstance_of_str(42)
     False
 
+    See also: ``i2.signatures.all_pk_signature``
+
     """
     from i2 import Sig, call_somewhat_forgivingly
 
@@ -890,51 +1017,95 @@ def nice_kinds(func):
 # wrap tools
 
 
+@double_up_as_factory
+def map_names(func=None, **old_to_new_name):
+    """Change the argument names of a function.
+
+    >>> def f(w, /, x: float, y=2, *, z: int = 3):
+    ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
+    >>> wrapped_f = map_names(f, w='DoubleYou', z='Zee')
+    >>> wrapped_f
+    <i2.Wrap f(DoubleYou, /, x: float, y=2, *, Zee: int = 3)>
+    >>> wrapped_f(1, 2, y=3, Zee=4)
+    '(w:=1) + (x:=2) * (y:=3) ** (z:=4) == 163'
+
+    Can also be used as a factory:
+    >>> @map_names(a='alpha', g='gamma')
+    ... def foo(a, b, g=1):
+    ...     return a + b * g
+    >>> foo(alpha=1, b=2, gamma=3)
+    7
+    """
+    return Ingress.name_map(func, **old_to_new_name).wrap(func)
+
+
 def include_exclude_ingress_factory(func, include=None, exclude=None):
     """A pattern underlying any ingress that takes a subset of parameters (possibly
     reordering them).
+
     For example: Keep only required arguments, or reorder params to be able to
     partialize #3 (without having to partialize #1 and #2)
+
     Note: A more general version would allow include and exclude to be expressed as
-    functions
-    that apply to one or several properties of the params (name, kind, default,
+    functions that apply to one or several properties of the params (name, kind, default,
     annotation).
     """
     sig = Sig(func)
+    if isinstance(include, str):
+        include = include.split()
+    if isinstance(exclude, str):
+        exclude = exclude.split()
     exclude = exclude or set()
     include = [x for x in (include or sig.names) if x not in exclude]
 
     return Ingress(inner_sig=sig, outer_sig=sig[include])
 
 
-def rm_params_ingress_factory(func, params_to_remove):
-    """Get a version of the function without some specific params"""
-    if isinstance(params_to_remove, str):
-        params_to_remove = params_to_remove.split()
-    return include_exclude_ingress_factory(func, exclude=params_to_remove)
+@double_up_as_factory
+def include_exclude(func=None, *, include=None, exclude=None):
+    """Reorder and/or remove parameters.
+
+    >>> def foo(a, b, c='C', d='D'):
+    ...     print(f"{a=},{b=},{c=},{d=}")
+    >>> bar = include_exclude(foo, include='b a', exclude='c d')
+
+    The signature of ``bar`` has only ``b`` and ``a``, in that order:
+
+    >>> from inspect import signature
+    >>> str(signature(bar))
+    '(b, a)'
+
+    But the function still works and does the same thing:
+
+    >>> bar('B', 'A')
+    a='A',b='B',c='C',d='D'
+
+    """
+    return wrap(func, ingress=include_exclude_ingress_factory(func, include, exclude))
 
 
-def rm_params(func, params_to_remove):
+@double_up_as_factory
+def rm_params(func=None, *, params_to_remove=()):
     """Get a function with some parameters removed.
 
     >>> from inspect import signature
     >>> def func(x, y=1, z=2):
     ...     return x + y * z
     >>>
-    >>> f = rm_params(func, 'z')
+    >>> f = rm_params(func, params_to_remove='z')
     >>> assert f(3) == func(3) == 5
     >>> assert f(3, 4) == func(3, 4) == 11
     >>> str(signature(f))
     '(x, y=1)'
     >>>
-    >>> f = rm_params(func, 'y z')
+    >>> f = rm_params(func, params_to_remove='y z')
     >>> assert f(3) == func(3) == 5
     >>> str(signature(f))
     '(x)'
-    
+
     But ``rm_params`` won't let you remove params that don't have defaults.
 
-    >>> f = rm_params(func, 'x z')
+    >>> f = rm_params(func, params_to_remove='x z')
     Traceback (most recent call last):
     ...
     AssertionError: Some of the params you want to remove don't have defaults: {'x'}
@@ -951,7 +1122,9 @@ def rm_params(func, params_to_remove):
         f'{params_to_remove_that_do_not_have_defaults}'
     )
 
-    return wrap(func, rm_params_ingress_factory(func, params_to_remove))
+    return wrap(
+        func, ingress=include_exclude_ingress_factory(func, exclude=params_to_remove)
+    )
 
 
 #     new_sig = sig - params_to_remove
@@ -1070,7 +1243,7 @@ def func_to_method_func(
     method_params=None,
     instance_arg_name='self',
 ) -> MethodFunc:
-    """Get a 'method function' from a 'normal function'.
+    """Get a 'method function' from a 'normal function'. Also known as "methodize".
 
     That is, get a function that gives the same outputs as the 'normal function',
     except that some of the arguments are sourced from the attributes of the first
@@ -1104,10 +1277,8 @@ def func_to_method_func(
     Now let's make a dummy object that has attributes ``a`` and a ``c``, and use it to
     call ``method_func``:
 
-    >>> class Klass:
-    ...     a = 1
-    ...     c = 3
-    >>> instance = Klass()
+    >>> from collections import namedtuple
+    >>> instance = namedtuple('FakeInstance', 'a c')(1, 3)
     >>> method_func(instance, 2, d='hello')
     'hello: 9'
 
@@ -1115,6 +1286,40 @@ def func_to_method_func(
 
     >>> assert method_func(instance, 2, d='hello') == func(1, 2, c=3, d='hello')
 
+    Often, though, what you'll want is to include this method function directly in
+    a class, as you're making that class "normally". That works too:
+
+    >>> from dataclasses import dataclass
+    >>> @dataclass
+    ... class Klass:
+    ...     a : int = 1
+    ...     c : int = 3
+    ...     method_func = func_to_method_func(func, 'a c')
+    >>> instance = Klass(1, 3)
+    >>> instance.method_func(2, d='hello')
+    'hello: 9'
+
+    What if your function has argument names that don't correspond to the names you
+    have, or want, as attributes of the class? Or even, you have several functions that
+    share an argument name that need to be bound to a different attribute?
+
+    For that, just use ``map_names`` to wrap the function, giving it the names that
+    you need to give it to have the effect you want (the binding of those arguments
+    to attributes of the instance):
+
+    >>> from i2.wrapper import map_names
+    >>> def func(x, y: int, z=2, *, d='bar'):
+    ...     return f"{d}: {(x + y) * z}"
+    >>> from dataclasses import dataclass
+    >>> @dataclass
+    ... class Klass:
+    ...     a : int = 1
+    ...     c : int = 3
+    ...     method_func = func_to_method_func(map_names(func, x='a', z='c'), 'a c')
+    >>> instance = Klass(1, 3)
+    >>> instance.method_func(2, d='hello')
+    'hello: 9'
+
     """
     # get a signature object for func
     sig = Sig(func)
@@ -1144,26 +1349,25 @@ def func_to_method_func(
         outer_sig=method_sig,  # this is the signature we want at the interface
     )
     # wrap the function, name it and return it
-    method_func = wrap(func, ingress)
+    method_func = wrap(func, ingress=ingress)
     method_func.__name__ = method_name
     return method_func
 
 
-from typing import Iterable
-
-
-def make_funcs_binding_class(
-    funcs, init_params=(), cls_name=None,
+def bind_funcs_object_attrs(
+    funcs, init_params=(), *, cls=None,
 ):
     """Transform one or several functions into a class that contains them as methods
     sourcing specific arguments from the instance's attributes.
 
     >>> from inspect import signature
+    >>> from dataclasses import dataclass
+    >>>
     >>> def foo(a, b, c=2, *, d='bar'):
     ...     return f"{d}: {(a + b) * c}"
     >>> foo(1, 2)
     'bar: 6'
-    >>> Klass = make_funcs_binding_class(foo, init_params='a c')
+    >>> Klass = bind_funcs_object_attrs(foo, init_params='a c')
     >>> Klass.__name__
     'Foo'
     >>> instance = Klass(a=1, c=3)
@@ -1181,26 +1385,478 @@ def make_funcs_binding_class(
     'hello: 9'
     >>> instance.foo(10, d='goodbye')
     'goodbye: 33'
+
+    >>> def foo(a, b, c):
+    ...     return a + b * c
+    ...
+    >>> def bar(d, e):
+    ...     return f"{d=}, {e=}"
+    ...
+    >>> @dataclass
+    ... class K:
+    ...     a: int
+    ...     e: int
+    ...
+    >>> C = bind_funcs_object_attrs([foo, bar], 'a e', cls=K)
+    >>> str(signature(C))
+    '(a: int, e: int) -> None'
+    >>> c = C(1,2)
+    >>> assert str(signature(c.foo)) == '(b, c)'
+    >>> c.foo(3,4)
+    13
+    >>> assert str(signature(c.bar)) == '(d)'
+    >>> c.bar(5)
+    'd=5, e=2'
     """
 
+    if isinstance(init_params, str):
+        init_params = init_params.split()
+
     dflt_cls_name = 'FuncsUnion'
     if callable(funcs) and not isinstance(funcs, Iterable):
         single_func = funcs
         dflt_cls_name = camelize(getattr(single_func, '__name__', dflt_cls_name))
         funcs = [single_func]
 
-    cls_name = cls_name or dflt_cls_name
-    if isinstance(init_params, str):
-        init_params = init_params.split()
+    if not isinstance(cls, type):
+        # if the class is not given, we need to make one
+        if isinstance(cls, str):
+            cls_name = cls
+        else:
+            cls_name = dflt_cls_name
 
-    # init_parameter_objects = Sig(func)[init_params].params
-    class_init_sig = Sig()
-    for func in funcs:
-        class_init_sig = class_init_sig.merge_with_sig(func)[init_params]
-    dataclass_fields = list(map(param_to_dataclass_field_tuple, class_init_sig.params))
-    Klass = make_dataclass(cls_name, dataclass_fields)
+        # init_parameter_objects = Sig(func)[init_params].params
+        # Make the signature for the init
+        class_init_sig = Sig()
+        for func in funcs:
+            class_init_sig = class_init_sig.merge_with_sig(func)[init_params]
+
+        dataclass_fields = list(
+            map(param_to_dataclass_field_tuple, class_init_sig.params)
+        )
+        cls = make_dataclass(cls_name, dataclass_fields)
 
     for func in funcs:
         method_func = func_to_method_func(func, init_params)
-        setattr(Klass, method_func.__name__, method_func)
-    return Klass
+        setattr(cls, method_func.__name__, method_func)
+    return cls
+
+
+def _items_filt(d: dict, keys):
+    for k, v in d.items():
+        if k in keys:
+            yield k, v
+
+
+def _mk_sig_from_params_and_funcs(params, funcs):
+    def gen():
+        for param in params:
+            pass
+
+
+## An attempt to redo a func_to_method_func because I forgot func_to_method_func existed
+## Has some different ideas, so keeping around until I decide it's time to let go.
+# NoSuchKey = type('NoSuchKey', (), {})
+# _instance_extractor: KwargsTrans
+#
+# # TODO: Add more (possibly optional) bind validation to fail early.
+# def _instance_extractor(
+#     outer_kwargs, bind: IdentifierMapping = (), instance_param: Identifier = 'self'
+# ):
+#     """
+#
+#     :param outer_kwargs: The input/outer keyword arguments
+#     :param bind: The inner->outer param mapping that defines what we want to extract
+#     :param instance_param: The name of the outer_kwargs key that contains the 'instance'
+#         from which we'll extract the bound
+#     :return:
+#     """
+#     """A KwargsTrans that extracts need arguments from one of the 'instance'
+#     outer_kwargs values.
+#
+#
+#     """
+#     # Compute the inverse {outer:inner,...} of {inner:outer,...} bind
+#     inv_bind = invert_map(bind)
+#     outer_kwargs = outer_kwargs.copy()
+#     instance = outer_kwargs.pop(instance_param)  # TODO: Better error handing
+#
+#     def gen():
+#         for outer_param, outer_val in outer_kwargs.items():
+#             if inner_param := inv_bind.get(outer_param, NoSuchKey) is not NoSuchKey:
+#                 # if outer_param was bound, the bound inner_param should be tied to
+#                 # the instance's attribute
+#                 yield inner_param, getattr(instance, outer_val)
+#             else:
+#                 # take the arg name and val as is
+#                 yield outer_param, outer_val
+#
+#     return dict(gen())
+
+#
+# def methodize(func, bind: Bind = ()):
+#     bind = identifier_mapping(bind)
+#     ingress = Ingress(
+#         outer_sig=Sig(func),
+#         kwargs_trans=partial(_instance_extractor, bind=bind),
+#         inner_sig=Sig('self') + Sig(func) - Sig(list(bind)),  # TODO: solve type or lint
+#     )
+#     return wrap(func, ingress=ingress)
+
+
+# ---------------------------------------------------------------------------------------
+# Extended Wrapper class
+
+from i2 import call_forgivingly
+
+
+class WrapperValidationError(ValueError):
+    """Raised when wrapper some construction params are not valid"""
+
+
+class EgressValidationError(WrapperValidationError):
+    """Raised when a egress is not valid"""
+
+
+class IngressValidationError(WrapperValidationError):
+    """Raised when a ingress is not valid"""
+
+
+class CallerValidationError(WrapperValidationError):
+    """Raised when a caller is not valid"""
+
+
+def _default_ingress(*args, **kwargs):
+    return args, kwargs
+
+
+def _default_egress(output, **egress_params):
+    return output
+
+
+def _default_caller(func, args, kwargs):
+    return func(*args, **kwargs)
+
+
+_keyword_kinds = {Sig.KEYWORD_ONLY, Sig.VAR_KEYWORD}
+
+
+def _all_kinds_are_keyword_only_or_variadic_keyword(sig):
+    return all(kind in _keyword_kinds for kind in list(sig.kinds.values())[3:])
+
+
+# TODO: Factor out more common parts with Wrap and reuse (possibly through _Wrap)
+class Wrapx(_Wrap):
+    def __init__(self, func, ingress=None, egress=None, *, caller=None, name=None):
+        """An extended wrapping object that allows more complex wrapping mechanisms.
+
+        :param func: The wrapped function
+        :param ingress: The incoming data transformer. It determines the argument properties
+            (name, kind, default and annotation) as well as the actual input of the
+            wrapped function.
+        :param egress: The outgoing data transformer. It also takes precedence over the
+            wrapped function to determine the return annotation of the ``Wrap`` instance
+        :param caller: A caller defines what it means to call the ``func`` on the
+            arguments it is given. It should be of the form
+            ``caller(func, args, kwargs, *, ...extra_keyword_only_params)``.
+            By default, the caller will simply return ``func(*args, **kwargs)``.
+        :param name: Name to give the wrapper (will use wrapped func name by default)
+
+        :return: A callable instance wrapping ``func``
+
+        >>> from inspect import signature
+        >>>
+        >>> def func(x, y):
+        ...     return x + y
+        ...
+        >>> def save_on_output_egress(v, *, k, s):
+        ...     s[k] = v
+        ...     return v
+        ...
+        >>> save_on_output = Wrapx(func, egress=save_on_output_egress)
+        >>> # TODO: should be `(x, y, *, k, s)` --> Need to work on the merge for this.
+        >>> str(signature(save_on_output))
+        '(x, y, k, s)'
+        >>>
+        >>> store = dict()
+        >>> save_on_output(1, 2, k='save_here', s=store)
+        3
+        >>> assert save_on_output(1, 2, k='save_here', s=store) == 3 == func(1, 2)
+        >>> store  # see what's in the store now!
+        {'save_here': 3}
+
+        A caller is meant to control the way the function is called.
+        It is given the ``func`` and the ``func_args`` and ``func_kwargs``
+        (whatever the ingress function gives it, if present) and possibly additional
+        params and will return... well, what ever you tell it to.
+
+        This can be used, for example, to call the function in a subprocess,
+        or on a remote system, differ computation (command pattern, for example, using
+        ``functools.partial``, or do what ever needs to have a view both on the function
+        and its inputs.
+
+        Here, we will wrap the function so it will apply to an iterable of inputs
+        (of the first argument), returning a list of results
+
+        >>> def func(x, y=2):
+        ...     return x + y
+        ...
+        >>> def iterize(func, args, kwargs):
+        ...     first_arg_val = next(iter(kwargs.values()))
+        ...     return list(map(func, first_arg_val))
+        ...
+        >>> iterized_func = Wrapx(func, caller=iterize)
+        >>> iterized_func([1, 2, 3, 4])
+        [3, 4, 5, 6]
+
+        Let's do the same as above, but allow other variables (here ``y``) to be input as
+        well. This takes a bit more work...
+
+        >>> from functools import partial
+        >>> def _iterize_first_arg(func, args, kwargs):
+        ...     first_arg_name = next(iter(kwargs))
+        ...     remaining_kwargs = {
+        ...         k: v for k, v in kwargs.items() if k != first_arg_name
+        ...     }
+        ...     return list(
+        ...         map(partial(func, **remaining_kwargs), kwargs[first_arg_name])
+        ...     )
+
+        Let's demo a different way of using Wrapx: Making a wrapper to apply at
+        function definition time
+
+        >>> iterize_first_arg = partial(Wrapx, caller=_iterize_first_arg)
+        >>> @iterize_first_arg
+        ... def func(x, y):
+        ...     return x + y
+        >>>
+        >>> func([1, 2, 3, 4], 10)
+        [11, 12, 13, 14]
+
+
+        """
+        super().__init__(func, ingress, egress, caller=caller, name=name)
+        self.ingress, self.egress, self.caller, self.sig = _process_wrapx_params(
+            func, ingress, egress, caller
+        )
+
+        self.__signature__ = self.sig
+        self.__wrapped__ = func
+        # TODO: Pros and cons analysis of pointing __wrapped__ to func. partial uses
+        #  .func, but wraps looks for __wrapped__
+
+    def __call__(self, *args, **kwargs):
+        try:
+            _kwargs = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
+            # TODO: Consider call_forgivingly(self.ingress, *args, **kwargs)
+            #  because call_forgivingly(self.ingress, **_kwargs) doesn't allow
+            #  same ingress functions to be used for Wrap and Wrapx
+            func_args, func_kwargs = call_forgivingly(self.ingress, **_kwargs)
+            inner_output = call_forgivingly(
+                self.caller, self.func, func_args, func_kwargs
+            )
+            return call_forgivingly(self.egress, inner_output, **_kwargs)
+        except Exception as e:
+            # Try running again, but with more careful validation, to try to give
+            # more specific error messages
+            # We don't do this in the first run so that we don't incur the validation
+            # overhead on every call.
+            _kwargs = self.sig.kwargs_from_args_and_kwargs(args, kwargs)
+            func_args, func_kwargs = _validate_ingress_output(
+                call_forgivingly(self.ingress, **_kwargs)
+            )
+            inner_output = call_forgivingly(self.func, *func_args, **func_kwargs)
+            return call_forgivingly(self.egress, inner_output, **_kwargs)
+
+
+def _validate_ingress_output(ingress_output):
+    if (
+        not isinstance(ingress_output, tuple)
+        or not len(ingress_output) == 2
+        or not isinstance(ingress_output[0], tuple)
+        or not isinstance(ingress_output[1], dict)
+    ):
+        raise IngressValidationError(
+            f'An ingress function should return a (tuple, dict) pair. '
+            f'This ingress function returned: {ingress_output}'
+        )
+    return ingress_output
+
+
+def _process_wrapx_params(func, ingress, egress, caller):
+    func_sig = Sig(func)
+
+    ingress, ingress_sig = _init_ingress(func_sig, ingress)
+    egress, egress_sig = _init_egress(func_sig, egress)
+    caller, caller_sig = _init_caller(caller)
+
+    sig = _mk_composite_sig(ingress_sig, egress_sig, caller_sig)
+    return ingress, egress, caller, sig
+
+
+def _mk_composite_sig(ingress_sig, egress_sig, caller_sig):
+    egress_sig_minus_first_arg = egress_sig - egress_sig.names[0]
+    caller_sig_minus_three_first_args = caller_sig - caller_sig.names[:3]
+    sig = Sig(
+        ingress_sig + egress_sig_minus_first_arg + caller_sig_minus_three_first_args,
+        return_annotation=egress_sig.return_annotation,
+    )
+    return sig
+
+
+def _init_caller(caller):
+    if caller is None:
+        caller = _default_caller
+        caller_sig = Sig('func args kwargs')  # sig with three inputs
+    else:
+        caller_sig = Sig(caller)
+        if len(caller_sig) < 3:
+            raise CallerValidationError(
+                f'A caller must have at least three arguments: '
+                f'{caller} signature was {caller_sig}'
+            )
+        if not _all_kinds_are_keyword_only_or_variadic_keyword(caller_sig):
+            raise CallerValidationError(
+                f'A caller must have at least three arguments'
+                f'{caller} signature was {caller_sig}'
+            )
+    return caller, caller_sig
+
+
+def _init_egress(func_sig, egress):
+    if egress is None:
+        egress = _default_egress
+        # signature with a single 'output' arg and func_sig's return_annotation
+        egress_sig = Sig('output', return_annotation=func_sig.return_annotation)
+        return_annotation = func_sig.return_annotation
+    else:
+        egress_sig = Sig(
+            egress,
+            # if egress has no return_annotation, use the func_sig's one.
+            # TODO: Is this really correct/safe? What if egress changes the type?
+            return_annotation=Sig(egress).return_annotation
+            or func_sig.return_annotation,
+        )
+
+    return egress, egress_sig
+
+
+def _init_ingress(func_sig, ingress):
+    if ingress is None:
+        ingress = _default_ingress
+        ingress_sig = func_sig
+    else:
+        ingress_sig = Sig(ingress)
+    return ingress, ingress_sig
+
+
+# ---------------------------------------------------------------------------------------
+# partialx
+
+
+def partialx(
+    func, *args, __name__=None, _rm_partialize=False, _allow_reordering=False, **kwargs
+):
+    """
+    Extends the functionality of builtin ``functools.partial`` with the ability to
+
+    - set ``__name__ ``
+
+    - remove partialized arguments from signature
+
+    - reorder params (so that defaults are at the end)
+
+    >>> def f(a, b=2, c=3):
+    ...     return a + b * c
+    >>> curried_f = partialx(f, c=10, _rm_partialize=True)
+    >>> curried_f.__name__
+    'f'
+    >>> from inspect import signature
+    >>> str(signature(curried_f))
+    '(a, b=2)'
+
+    >>> def f(a, b, c=3):
+    ...     return a + b * c
+
+    Note that ``a`` gets a default, but ``b`` does not, yet is after ``a``.
+    This is allowed because these parameters all became KEYWORD_ONLY.
+
+    >>> g = partialx(f, a=1)
+    >>> str(Sig(g))
+    '(*, a=1, b, c=3)'
+
+    If you wanted to reorder the parameters to have all defaulted kinds be at the end,
+    as usual, you can do so using ``_allow_reordering=True``
+
+    >>> g = partialx(f, a=1, _allow_reordering=True)
+    >>> str(Sig(g))
+    '(*, b, a=1, c=3)'
+
+    """
+    f = partial(func, *args, **kwargs)
+    if _rm_partialize:
+        sig = Sig(func)
+        partialized = list(
+            sig.kwargs_from_args_and_kwargs(args, kwargs, allow_partial=True)
+        )
+        sig = sig - partialized
+        f = sig(partial(f, *args, **kwargs))
+    if _allow_reordering:
+        # TODO: Instead of Sig(f).defaults, move only params that need to move
+        # TODO: + Change signature so that the number of params that become keyword-only
+        #   is minimize.
+        f = move_params_to_the_end(f, Sig(f).defaults)
+    f.__name__ = __name__ or name_of_obj(func)
+    return f
+
+
+def move_params_to_the_end(
+    func: Callable, names_to_move: Union[Callable, Iterable[str]]
+):
+    """
+    Choose args from func, according to choice_args_func and move them
+    to the right
+
+    >>> from functools import partial
+    >>> from i2 import Sig
+    >>> def foo(a, b, c):
+    ...     return a + b + c
+    >>> g = partial(foo, b=4)  # fixing a, which is before b
+    >>> h = move_params_to_the_end(g, Sig(g).defaults)
+    >>> assert str(Sig(g)) == '(a, *, b=4, c)'
+    >>> assert str(Sig(h)) == '(a, *, c, b=4)'
+
+    """
+    if callable(names_to_move):
+        names_to_move = names_to_move(func)
+    assert isinstance(names_to_move, Iterable), (
+        f'names_to_move must be an iterable of names '
+        f'or a callable producing one from a function. Was {names_to_move}'
+    )
+
+    names = Sig(func).names
+    reordered = move_names_to_the_end(names, names_to_move)
+    wrapped_func = include_exclude(func, include=reordered)
+    return wrapped_func
+
+
+def move_names_to_the_end(names, names_to_move_to_the_end):
+    """
+    Remove the items of ``names_to_move_to_the_end`` from ``names``
+    and append to the right of names
+
+    >>> names = ['a','c','d','e']
+    >>> names_to_move_to_the_end = ['c','e']
+    >>> move_names_to_the_end(names, names_to_move_to_the_end)
+    ['a', 'd', 'c', 'e']
+    >>> names_to_move_to_the_end = 'c e'
+    >>> move_names_to_the_end(names, names_to_move_to_the_end)
+    ['a', 'd', 'c', 'e']
+
+    """
+    if isinstance(names_to_move_to_the_end, str):
+        names_to_move_to_the_end = names_to_move_to_the_end.split()
+    else:
+        names_to_move_to_the_end = list(names_to_move_to_the_end)
+    removed = [x for x in names if x not in names_to_move_to_the_end]
+    return list(removed) + names_to_move_to_the_end
